"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["ResponseFuture", "Handler", "ThreadingHandler", "RequestHandler"]
import atexit
import threading
import Queue

from collections import namedtuple


class ResponseFuture(object):
    """A response which may have a value at some point."""

    def __init__(self, handler):
        """
        :type handler: :class:`pykafka.handlers.Handler`
        """
        self.handler = handler
        self.error = False
        self._ready = handler.Event()

    def set_response(self, response):
        """Set response data and trigger get method."""
        self.response = response
        self._ready.set()

    def set_error(self, error):
        """Set error and trigger get method."""
        self.error = error
        self._ready.set()

    def get(self, response_cls=None, timeout=None):
        """Block until data is ready and return.

        Raises an exception if there was an error.
        """
        self._ready.wait(timeout)
        if self.error:
            raise self.error
        if response_cls:
            return response_cls(self.response)
        else:
            return self.response


class Handler(object):
    """Base class for Handler classes"""
    def spawn(self, target, *args, **kwargs):
        """Create the worker that will process the work to be handled"""
        raise NotImplementedError


class ThreadingHandler(Handler):
    """A handler. that uses a :class:`threading.Thread` to perform its work"""
    QueueEmptyError = Queue.Empty
    Queue = Queue.Queue
    Event = threading.Event
    Lock = threading.Lock

    def spawn(self, target, *args, **kwargs):
        t = threading.Thread(target=target, *args, **kwargs)
        t.daemon = True
        t.start()
        return t


class RequestHandler(object):
    """Uses a Handler instance to dispatch requests."""

    Task = namedtuple('Task', ['request', 'future'])

    def __init__(self, handler, connection):
        """
        :type handler: :class:`pykafka.handlers.Handler`
        :type connection: :class:`pykafka.connection.BrokerConnection`
        """
        self.handler = handler
        self.connection = connection
        self._requests = handler.Queue()
        self.ending = handler.Event()
        atexit.register(self.stop)

    def request(self, request, has_response=True):
        """Construct a new request

        :type request: :class:`pykafka.protocol.Request`
        :param has_response: Whether this request will return a response
        :returns: :class:`pykafka.handlers.ResponseFuture`
        """
        future = None
        if has_response:
            future = ResponseFuture(self.handler)

        task = self.Task(request, future)
        self._requests.put(task)
        return future

    def start(self):
        """Start the request processor."""
        self.t = self._start_thread()

    def stop(self):
        """Stop the request processor."""
        self._requests.join()
        self.ending.set()

    def _start_thread(self):
        """Run the request processor"""
        def worker():
            while not self.ending.is_set():
                task = self._requests.get()
                try:
                    self.connection.request(task.request)
                    if task.future:
                        res = self.connection.response()
                        task.future.set_response(res)
                except Exception, e:
                    if task.future:
                        task.future.set_error(e)
                finally:
                    self._requests.task_done()
        return self.handler.spawn(worker)



from __future__ import division
"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Cluster"]
import logging
import random
import time
import weakref

from .broker import Broker
from .exceptions import (ConsumerCoordinatorNotAvailable,
                         KafkaException,
                         UnknownTopicOrPartition)
from .protocol import ConsumerMetadataRequest, ConsumerMetadataResponse
from .topic import Topic


log = logging.getLogger(__name__)


class TopicDict(dict):
    """Dictionary which will attempt to auto-create unknown topics."""

    def __init__(self, cluster, *args, **kwargs):
        super(TopicDict, self).__init__(*args, **kwargs)
        self._cluster = weakref.proxy(cluster)

    def __missing__(self, key):
        log.warning('Topic %s not found. Attempting to auto-create.', key)
        if self._create_topic(key):
            return self[key]
        else:
            raise UnknownTopicOrPartition('Unknown topic: {}'.format(key))

    def _create_topic(self, topic_name):
        """Auto-create a topic.

        Not exposed in the cluster or broker because this is *only*
        auto-creation.  When there's a real API for creating topics,
        with settings and everything, we'll implement that. To expose just
        this now would be disingenuous, since it's features would be hobbled.
        """
        if len(self._cluster.brokers) == 0:
            log.warning("No brokers found. This is probably because of "
                        "KAFKA-2154, which will be fixed in Kafka 0.8.3")
            raise KafkaException("Unable to retrieve metdata. Can't auto-create topic. See log for details.")
        # Auto-creating will take a moment, so we try 5 times.
        for i in xrange(5):
            # Auto-creating is as simple as issuing a metadata request
            # solely for that topic.  The update is just to be sure
            # our `Cluster` knows about it.
            self._cluster.brokers[self._cluster.brokers.keys()[0]].request_metadata(topics=[topic_name])
            self._cluster.update()
            if topic_name in self:
                log.info('Topic %s successfully auto-created.', topic_name)
                return True
            time.sleep(0.1)


class Cluster(object):
    """
    A Cluster is a high-level abstraction of the collection of brokers and
    topics that makes up a real kafka cluster.
    """
    def __init__(self,
                 hosts,
                 handler,
                 socket_timeout_ms=30 * 1000,
                 offsets_channel_socket_timeout_ms=10 * 1000,
                 exclude_internal_topics=True):
        """Create a new Cluster instance.

        :param hosts: Comma-separated list of kafka hosts to used to connect.
        :type hosts: str
        :param handler: The concurrency handler for network requests.
        :type handler: :class:`pykafka.handlers.Handler`
        :param socket_timeout_ms: The socket timeout (in milliseconds) for
            network requests
        :type socket_timeout_ms: int
        :param offsets_channel_socket_timeout_ms: The socket timeout (in
            milliseconds) when reading responses for offset commit and
            offset fetch requests.
        :type offsets_channel_socket_timeout_ms: int
        :param exclude_internal_topics: Whether messages from internal topics
            (specifically, the offsets topic) should be exposed to consumers.
        :type exclude_internal_topics: bool
        """
        self._seed_hosts = hosts
        self._socket_timeout_ms = socket_timeout_ms
        self._offsets_channel_socket_timeout_ms = offsets_channel_socket_timeout_ms
        self._handler = handler
        self._brokers = {}
        self._topics = TopicDict(self)
        self._exclude_internal_topics = exclude_internal_topics
        self.update()

    def __repr__(self):
        return "<{}.{} at {} (hosts={})>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self._seed_hosts,
        )

    @property
    def brokers(self):
        """The dict of known brokers for this cluster"""
        return self._brokers

    @property
    def topics(self):
        """The dict of known topics for this cluster"""
        return self._topics

    @property
    def handler(self):
        """The concurrency handler for network requests"""
        return self._handler

    def _get_metadata(self):
        """Get fresh cluster metadata from a broker."""
        # Works either on existing brokers or seed_hosts list
        brokers = [b for b in self.brokers.values() if b.connected]
        if brokers:
            for broker in brokers:
                response = broker.request_metadata()
                if response is not None:
                    return response
        else:  # try seed hosts
            brokers = self._seed_hosts.split(',')
            for broker_str in brokers:
                try:
                    h, p = broker_str.split(':')
                    broker = Broker(-1, h, p, self._handler,
                                    self._socket_timeout_ms,
                                    self._offsets_channel_socket_timeout_ms,
                                    buffer_size=1024 * 1024)
                    response = broker.request_metadata()
                    if response is not None:
                        return response
                except:
                    log.exception('Unable to connect to broker %s',
                                  broker_str)
        # Couldn't connect anywhere. Raise an error.
        raise Exception('Unable to connect to a broker to fetch metadata.')

    def _update_brokers(self, broker_metadata):
        """Update brokers with fresh metadata.

        :param broker_metadata: Metadata for all brokers.
        :type broker_metadata: Dict of `{name: metadata}` where `metadata` is
            :class:`pykafka.protocol.BrokerMetadata` and `name` is str.
        """
        # FIXME: A cluster with no topics returns no brokers in metadata
        # Remove old brokers
        removed = set(self._brokers.keys()) - set(broker_metadata.keys())
        if len(removed) > 0:
            log.info('Removing %d brokers', len(removed))
        for id_ in removed:
            log.debug('Removing broker %s', self._brokers[id_])
            self._brokers.pop(id_)
        # Add/update current brokers
        if len(broker_metadata) > 0:
            log.info('Discovered %d brokers', len(broker_metadata))
        for id_, meta in broker_metadata.iteritems():
            if id_ not in self._brokers:
                log.debug('Discovered broker id %s: %s:%s', id_, meta.host, meta.port)
                self._brokers[id_] = Broker.from_metadata(
                    meta, self._handler, self._socket_timeout_ms,
                    self._offsets_channel_socket_timeout_ms,
                    buffer_size=1024 * 1024
                )
            else:
                broker = self._brokers[id_]
                if meta.host == broker.host and meta.port == broker.port:
                    continue  # no changes
                # TODO: Can brokers update? Seems like a problem if so.
                #       Figure out and implement update/disconnect/reconnect if
                #       needed.
                raise Exception('Broker host/port change detected! %s', broker)

    def _update_topics(self, metadata):
        """Update topics with fresh metadata.

        :param metadata: Metadata for all topics.
        :type metadata: Dict of `{name, metadata}` where `metadata` is
            :class:`pykafka.protocol.TopicMetadata` and `name` is str.
        """
        # Remove old topics
        removed = set(self._topics.keys()) - set(metadata.keys())
        if len(removed) > 0:
            log.info("Removing %d topics", len(removed))
        for name in removed:
            log.debug('Removing topic %s', self._topics[name])
            self._topics.pop(name)
        # Add/update partition information
        if len(metadata) > 0:
            log.info("Discovered %d topics", len(metadata))
        for name, meta in metadata.iteritems():
            if not self._should_exclude_topic(name):
                if name not in self._topics:
                    self._topics[name] = Topic(self, meta)
                    log.debug('Discovered topic %s', self._topics[name])
                else:
                    self._topics[name].update(meta)

    def _should_exclude_topic(self, topic_name):
        """Should this topic be excluded from the list shown to the client?"""
        if not self._exclude_internal_topics:
            return False
        return topic_name.startswith("__")

    def get_offset_manager(self, consumer_group):
        """Get the broker designated as the offset manager for this consumer group.

        Based on Step 1 at https://cwiki.apache.org/confluence/display/KAFKA/Committing+and+fetching+consumer+offsets+in+Kafka

        :param consumer_group: The name of the consumer group for which to
            find the offset manager.
        :type consumer_group: str
        """
        log.info("Attempting to discover offset manager for consumer group '%s'",
                 consumer_group)
        # arbitrarily choose a broker, since this request can go to any
        broker = self.brokers[random.choice(self.brokers.keys())]
        MAX_RETRIES = 5

        for i in xrange(MAX_RETRIES):
            if i > 0:
                log.debug("Retrying offset manager discovery")
            time.sleep(i * 2)

            req = ConsumerMetadataRequest(consumer_group)
            future = broker.handler.request(req)
            try:
                res = future.get(ConsumerMetadataResponse)
            except ConsumerCoordinatorNotAvailable:
                log.error('Error discovering offset manager.')
                if i == MAX_RETRIES - 1:
                    raise
            else:
                coordinator = self.brokers.get(res.coordinator_id, None)
                if coordinator is None:
                    raise Exception('Coordinator broker with id {} not found'.format(res.coordinator_id))
                log.info("Found coordinator broker with id %s", res.coordinator_id)
                return coordinator

    def update(self):
        """Update known brokers and topics."""
        metadata = self._get_metadata()
        if len(metadata.brokers) == 0 and len(metadata.topics) == 0:
            log.warning('No broker metadata found. If this is a fresh cluster, '
                        'this may be due to a bug in Kafka. You can force '
                        'broker metadata to be returned by manually creating '
                        'a topic in the cluster. See '
                        'https://issues.apache.org/jira/browse/KAFKA-2154 '
                        'for information. Please note: topic auto-creation '
                        'will NOT work. You need to create at least one topic '
                        'manually using the Kafka CLI tools.')
        self._update_brokers(metadata.brokers)
        self._update_topics(metadata.topics)



"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["BaseCluster", "BaseBroker", "BasePartition", "BaseTopic",
           "BaseSimpleConsumer", "BaseProducer", "BaseAsyncProducer"]
from common import CompressionType
from partitioners import random_partitioner


class BaseCluster(object):
    """A Kafka cluster.

    This is an abstraction of the cluster topology. It provides access
    to topics and brokers, which can be useful for introspection of a cluster.
    """

    @property
    def brokers(self):
        """Brokers associated with this cluster.

        :type: `dict` of {broker_id: :class:`pykafka.base.BaseBroker`}
        """
        return self._brokers

    @property
    def topics(self):
        """Topics present in this cluster.

        :type: `dict` of {topic_name: :class:`pykafka.base.BaseTopic`}
        """
        return self._topics

    def update(self):
        """Update the Cluster with metadata from Kafka.

        All updates must happen in-place. This means that if a Topic leader has
        changed, a new Topic can't be created and put into `self.topics`. That
        would break any clients that have instances of the old Topic. Instead,
        the current topic is updated seamlessly.
        """
        raise NotImplementedError


class BaseBroker(object):
    """A Kafka Broker.

    Not especially useful under normal circumstances, but can be handy
    when introspecting about a Cluster.
    """

    @property
    def id(self):
        """Id of the broker

        :type: `int`
        """
        return self._id

    @property
    def host(self):
        """Host of the broker.

        :type: `str`
        """
        return self._host

    @property
    def port(self):
        """Port the broker uses.

        :type: `int`
        """
        return self._port


class BasePartition(object):
    """A Kafka Partition.

    Each Kafka topic is split up into parts called "partitions".  When reading
    or writing a topic, you're actually reading a partition of the topic.
    Replication also happens at the partition level.

    Like Brokers, Partitions aren't useful under normal circumstances, but
    are handy to know about for debugging and introspection.
    """

    @property
    def id(self):
        """The id of this partition.

        :type: `int`
        """
        return self._id

    @property
    def leader(self):
        """The leader broker of the partition.

        :type: :class:`pykafka.base.BasePartition`
        """
        return self._leader

    @property
    def replicas(self):
        """List of brokers which has replicas of this Partition.

        :type: `list` of :class:`pykafka.base.BaseBroker`
        """
        return self._replicas

    @property
    def isr(self):
        """List of brokers which have in-sync replicas of this partition.

        :type: `list` of :class:`pykafka.base.BaseBroker`
        """
        return self._isr

    @property
    def topic(self):
        """Name of the topic to which this partition belongs.

        :type: `str`
        """
        return self._topic

    def latest_available_offsets(self):
        """Gets the latest offset for the partition.

        :return: Latest offset for the partition.
        :rtype: `int`
        """
        raise NotImplementedError

    def earliest_available_offsets(self):
        """Gets the earliest offset for the partition.

        Due to logfile rotation, this will not always be 0. Instead,
        this will get the earliest offset for which the partition has data.

        :returns: The earliest offset for the partition.
        :rtype: `int`
        """
        raise NotImplementedError


class BaseTopic(object):
    """A Kafka topic."""

    @property
    def name(self):
        """The name of the topic.

        :type: `str`
        """
        return self._name

    @property
    def partitions(self):
        """The partitions of this topic.

        :type: `dict` of {`int`: :class:`pykafka.base.BasePartition`}
        """
        return self._partitions

    def latest_offsets(self):
        """Get the latest offset for all partitions.

        :returns: The latest offset for all partitions in the topic.
        :rtype: `dict` of {:class:`pykafka.base.BasePartition`: `int`}
        """
        raise NotImplementedError

    def earliest_offsets(self):
        """Get the earliest offset for all partitions.

        Due to logfile rotation, this will not always be 0. Instead,
        this will get the earliest offset for which the partition has data.

        :returns: The earliest offset for all partitions in the topic.
        :rtype: `dict` of {:class:`pykafka.base.BasePartition`: `int`}
        """
        raise NotImplementedError


class BaseSimpleConsumer(object):
    """A simple consumer which reads data from a topic.

    This is a simple consumer useful for testing or single-process situations.
    **If multiple processes use a SimpleConsumer and read the same topic,
    they will each read copies of the data.**  Instead, use a BalancedConsumer
    to ensure each message is only read once.

    The one advantage this implementation has over a BalancedConsumer is that
    the partitions to be read can be specified. Therefore, if one has hard
    coded which processes reads which partitions, this is a useful soluton.
    """

    def __init__(self, client, topic, partitions=None):
        """Create a consumer for a topic.

        :param client: Client connection to the cluster.
        :type client: :class:`pykafka.client.KafkaClient`
        :param topic: The topic to consume from.
        :type topic: :class:`pykafka.base.BaseTopic` or :class:`str`
        :param partitions: List of partitions to consume from.
        :type partitions: Iterable of :class:`pykafka.base.BasePartition` or int
        """
        raise NotImplementedError

    def __iter__(self):
        """Iterator for messages in the consumer."""
        raise NotImplementedError

    @property
    def topic(self):
        """The topic from which data is being read.

        :type: :class:`pykafka.base.BaseTopic`
        """
        return self._topic

    @property
    def partitions(self):
        """The partitions from which data is being read.

        :type: `dict` of {`int`: :class:`kafka.base.BasePartition`}
        """
        return self._partitions

    def consume(self, timeout=None):
        """Consume a message from the topic.

        :returns: A message.
        :rtype: :class:`kafka.common.Message`
        """
        raise NotImplementedError


class BaseProducer(object):
    """A producer which writes data to a topic.

    This producer is synchronous, waiting for a response from Kafka
    before returning. For an asynchronous implementation, use
    :class:`pykafka.base.BaseAsyncProducer`
    """
    def __init__(self,
                 client,
                 topic,
                 partitioner=random_partitioner,
                 compression=CompressionType.NONE,
                 max_retries=3,
                 retry_backoff_ms=100,
                 topic_refresh_interval_ms=600000,
                 required_acks=1,
                 ack_timeout_ms=10000,
                 batch_size=200):
        """Create a Producer for a topic.

        :param client: KafkaClient used to connect to cluster.
        :param topic: The topic to produce messages for.
        :type topic: :class:`pykafka.topic.Topic`
        :para compression: Compression to use for messages.
        :type compression: :class:`kafka.common.CompressionType`
        :param max_retries: Number of times to retry sending messages.
        :param retry_backoff_ms: Interval (in milliseconds) to wait between
            retries
        :param topic_refresh_interval_ms: Time (in milliseconds) between queries
            to refresh metadata about the topic. The Producer will also refresh
            this data when the cluster changes (e.g. partitions missing, etc),
            but this is the interval for how often it actively polls for
            changes.
        :param required_acks: How many other brokers must have committed the
            data to their log and acknowledged this to the leader before a
            request is considered complete?
        :param ack_timeout_ms: Amount of time (in milliseconds) to wait for
            acknowledgment of a produce request.
        :param batch_size: Size (in bytes) of batches to send to brokers.
        """
        raise NotImplementedError

    @property
    def topic(self):
        """The topic to which data is being written.

        :type: :class:`pykafka.base.BaseTopic`
        """
        return self._topic

    @property
    def partitioner(self):
        """The partitioner used to determine which partition used.

        :type: :class:`pykafka.partitioners.BasePartitioner`
        """
        return self._partitioner

    def produce(self, messages):
        """Produce messages to the topic.

        :param messages: Iterable of messages to be published.
        :type messages: An iterable of either strings or (key, value) tuples.
            If tuples, then the `key` will be sent to the partitioner to
            determine to which partition it belongs. It will also be sent to
            Kafka and available when the message is read.
        """
        raise NotImplementedError


class BaseAsyncProducer(BaseProducer):
    """Asynchronous Producer for the Kafka Cluster.

    Asynchronously publishes messages to the Kafka cluster. Calling `produce`
    will return immediately.  Messages will be batched and published at regular
    intervals based on settings passed to the AsyncProducer.
    """
    pass



import math
from uuid import uuid4

import mock
import unittest2

from pykafka import KafkaClient
from pykafka.balancedconsumer import BalancedConsumer
from pykafka.test.utils import get_cluster, stop_cluster


class TestBalancedConsumer(unittest2.TestCase):
    def test_decide_partitions(self):
        """Test partition assignment for a number of partitions/consumers."""
        consumer_group = 'testgroup'
        # 100 test iterations
        for i in xrange(100):
            # Set up partitions, cluster, etc
            num_participants = i + 1
            num_partitions = 100 - i
            participants = ['test-debian:{}'.format(p)
                            for p in xrange(num_participants)]
            topic = mock.Mock()
            topic.name = 'testtopic'
            topic.partitions = {}
            for k in xrange(num_partitions):
                part = mock.Mock(name='part-{}'.format(k))
                part.id = k
                part.topic = topic
                part.leader = mock.Mock()
                part.leader.id = k % num_participants
                topic.partitions[k] = part

            cluster = mock.MagicMock()
            zk = mock.MagicMock()
            cns = BalancedConsumer(topic, cluster, consumer_group,
                                   zookeeper=zk, auto_start=False)

            # Simulate each participant to ensure they're correct
            assigned_parts = []
            for p_id in xrange(num_participants):
                cns._consumer_id = participants[p_id]  # override consumer id

                # Decide partitions then validate
                partitions = cns._decide_partitions(participants)
                assigned_parts.extend(partitions)

                remainder_ppc = num_partitions % num_participants
                idx = participants.index(cns._consumer_id)
                parts_per_consumer = num_partitions / num_participants
                parts_per_consumer = math.floor(parts_per_consumer)
                num_parts = parts_per_consumer + (0 if (idx + 1 > remainder_ppc) else 1)

                self.assertEqual(len(partitions), num_parts)

            # Validate all partitions were assigned once and only once
            all_partitions = topic.partitions.values()
            all_partitions.sort()
            assigned_parts.sort()
            self.assertListEqual(assigned_parts, all_partitions)


class BalancedConsumerIntegrationTests(unittest2.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = 'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.client = KafkaClient(cls.kafka.brokers)
        prod = cls.client.topics[cls.topic_name].get_producer(batch_size=5)
        prod.produce('msg {}'.format(i) for i in xrange(1000))

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_consume(self):
        try:
            consumer_a = self.client.topics[self.topic_name].get_balanced_consumer('test_consume', zookeeper_connect=self.kafka.zookeeper)
            consumer_b = self.client.topics[self.topic_name].get_balanced_consumer('test_consume', zookeeper_connect=self.kafka.zookeeper)

            # Consume from both a few times
            messages = [consumer_a.consume() for i in xrange(1)]
            self.assertTrue(len(messages) == 1)
            messages = [consumer_b.consume() for i in xrange(1)]
            self.assertTrue(len(messages) == 1)

            # Validate they aren't sharing partitions
            self.assertSetEqual(
                consumer_a._partitions & consumer_b._partitions,
                set()
            )

            # Validate all partitions are here
            self.assertSetEqual(
                consumer_a._partitions | consumer_b._partitions,
                set(self.client.topics[self.topic_name].partitions.values())
            )
        finally:
            consumer_a.stop()
            consumer_b.stop()



if __name__ == "__main__":
    unittest2.main()



from __future__ import division
"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Producer", "AsyncProducer"]
import itertools
import logging
import time
from collections import defaultdict

import base
from .common import CompressionType
from .exceptions import (
    UnknownTopicOrPartition, NotLeaderForPartition, RequestTimedOut,
    ProduceFailureError, SocketDisconnectedError, InvalidMessageError,
    InvalidMessageSize, MessageSizeTooLarge
)
from .partitioners import random_partitioner
from .protocol import Message, ProduceRequest


log = logging.getLogger(__name__)


class AsyncProducer(base.BaseAsyncProducer):
    def __init__(self,
                 topic,
                 partitioner=None,
                 compression=CompressionType.NONE,
                 max_retries=3,
                 retry_backoff_ms=100,
                 topic_refresh_interval_ms=600000,
                 required_acks=0,
                 ack_timeout_ms=10000,
                 batch_size=200,
                 batch_time_ms=5000,
                 max_pending_messages=10000):
        """Create an AsyncProducer for a topic."""
        raise NotImplementedError("AsyncProducer is unimplemented")


class Producer(base.BaseProducer):
    """
    This class implements the synchronous producer logic found in the
    JVM driver.
    """
    def __init__(self,
                 cluster,
                 topic,
                 partitioner=random_partitioner,
                 compression=CompressionType.NONE,
                 max_retries=3,
                 retry_backoff_ms=100,
                 topic_refresh_interval_ms=600000,
                 required_acks=1,
                 ack_timeout_ms=10000,
                 batch_size=200):
        """Instantiate a new Producer.

        :param cluster: The cluster to which to connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param topic: The topic to which to produce messages
        :type topic: :class:`pykafka.topic.Topic`
        :param partitioner: The partitioner to use during message production
        :type partitioner: :class:`pykafka.partitioners.BasePartitioner`
        :param compression: The type of compression to use.
        :type compression: :class:`pykafka.common.CompressionType`
        :param max_retries: How many times to attempt to produce messages
            before raising an error.
        :type max_retries: int
        :param retry_backoff_ms: The amount of time (in milliseconds) to
            back off during produce request retries.
        :type retry_backoff_ms: int
        :param topic_refresh_interval_ms: Time (in milliseconds) between queries
            to refresh metadata about the topic. The Producer will also refresh
            this data when the cluster changes (e.g. partitions missing, etc),
            but this is the interval for how often it actively polls for
            changes.
        :type topic_refresh_interval_ms: int
        :param required_acks: How many other brokers must have committed the
            data to their log and acknowledged this to the leader before a
            request is considered complete?
        :type required_acks: int
        :param ack_timeout_ms: Amount of time (in milliseconds) to wait for
            acknowledgment of a produce request.
        :type ack_timeout_ms: int
        :param batch_size: Size (in bytes) of batches to send to brokers.
        :type batch_size: int
        """
        # See BaseProduce.__init__.__doc__ for docstring
        self._cluster = cluster
        self._topic = topic
        self._partitioner = partitioner
        self._compression = compression
        self._max_retries = max_retries
        self._retry_backoff_ms = retry_backoff_ms
        self._required_acks = required_acks
        self._ack_timeout_ms = ack_timeout_ms
        self._batch_size = batch_size

    def __repr__(self):
        return "<{}.{} at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self))
        )

    def _send_request(self, broker, req, attempt):
        """Send the produce request to the broker and handle the response.

        :param broker: The broker to which to send the request
        :type broker: :class:`pykafka.broker.Broker`
        :param req: The produce request to send
        :type req: :class:`pykafka.protocol.ProduceRequest`
        :param attempt: The current attempt count. Used for retry logic
        :type attempt: int
        """

        def _get_partition_msgs(partition_id, req):
            """Get all the messages for the partitions from the request."""
            messages = itertools.chain.from_iterable(
                mset.messages
                for topic, partitions in req.msets.iteritems()
                for p_id, mset in partitions.iteritems()
                if p_id == partition_id
            )
            for message in messages:
                yield (message.partition_key, message.value), partition_id

        # Do the request
        to_retry = []
        try:
            response = broker.produce_messages(req)

            # Figure out if we need to retry any messages
            # TODO: Convert to using utils.handle_partition_responses
            to_retry = []
            for topic, partitions in response.topics.iteritems():
                for partition, presponse in partitions.iteritems():
                    if presponse.err == 0:
                        continue  # All's well
                    if presponse.err == UnknownTopicOrPartition.ERROR_CODE:
                        log.warning('Unknown topic: %s or partition: %s. '
                                    'Retrying.', topic, partition)
                    elif presponse.err == NotLeaderForPartition.ERROR_CODE:
                        log.warning('Partition leader for %s/%s changed. '
                                    'Retrying.', topic, partition)
                        # Update cluster metadata to get new leader
                        self._cluster.update()
                    elif presponse.err == RequestTimedOut.ERROR_CODE:
                        log.warning('Produce request to %s:%s timed out. '
                                    'Retrying.', broker.host, broker.port)
                    elif presponse.err == InvalidMessageError.ERROR_CODE:
                        log.warning('Encountered InvalidMessageError')
                    elif presponse.err == InvalidMessageSize.ERROR_CODE:
                        log.warning('Encountered InvalidMessageSize')
                        continue
                    elif presponse.err == MessageSizeTooLarge.ERROR_CODE:
                        log.warning('Encountered MessageSizeTooLarge')
                        continue
                    to_retry.extend(_get_partition_msgs(partition, req))
        except SocketDisconnectedError:
            log.warning('Broker %s:%s disconnected. Retrying.',
                        broker.host, broker.port)
            self._cluster.update()
            to_retry = [
                ((message.partition_key, message.value), p_id)
                for topic, partitions in req.msets.iteritems()
                for p_id, mset in partitions.iteritems()
                for message in mset.messages
            ]

        if to_retry:
            attempt += 1
            if attempt < self._max_retries:
                time.sleep(self._retry_backoff_ms / 1000)
                self._produce(to_retry, attempt)
            else:
                raise ProduceFailureError('Unable to produce messages. See log for details.')

    def _partition_messages(self, messages):
        """Assign messages to partitions using the partitioner.

        :param messages: Iterable of messages to publish.
        :returns:        Generator of ((key, value), partition_id)
        """
        partitions = self._topic.partitions.values()
        for message in messages:
            if isinstance(message, basestring):
                key = None
                value = message
            else:
                key, value = message
            value = str(value)
            yield (key, value), self._partitioner(partitions, message).id

    def _produce(self, message_partition_tups, attempt):
        """Publish a set of messages to relevant brokers.

        :param message_partition_tups: Messages with partitions assigned.
        :type message_partition_tups:  tuples of ((key, value), partition_id)
        """
        # Requests grouped by broker
        requests = defaultdict(lambda: ProduceRequest(
            compression_type=self._compression,
            required_acks=self._required_acks,
            timeout=self._ack_timeout_ms,
        ))

        for ((key, value), partition_id) in message_partition_tups:
            # N.B. This handles retries, so the leader lookup is needed
            leader = self._topic.partitions[partition_id].leader
            requests[leader].add_message(
                Message(value, partition_key=key),
                self._topic.name,
                partition_id
            )
            # Send requests at the batch size
            if requests[leader].message_count() >= self._batch_size:
                self._send_request(leader,
                                   requests.pop(leader),
                                   attempt)

        # Send any still not sent
        for leader, req in requests.iteritems():
            self._send_request(leader, req, attempt)

    def produce(self, messages):
        """Produce a set of messages.

        :param messages: The messages to produce
        :type messages: Iterable of str or (str, str) tuples
        """
        # Do partition distribution here. We need to be able to retry producing
        # only *some* messages when a leader changes. Therefore, we don't want
        # a random partition distribution changing that on the retry.
        self._produce(self._partition_messages(messages), 0)



import mock
import os
import time
import unittest2

from pykafka import KafkaClient
from pykafka.simpleconsumer import OwnedPartition
from pykafka.test.utils import get_cluster, stop_cluster


class TestSimpleConsumer(unittest2.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = 'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.kafka.produce_messages(
            cls.topic_name,
            ('msg {}'.format(i) for i in xrange(1000))
        )
        cls.client = KafkaClient(cls.kafka.brokers)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_consume(self):
        consumer = self.client.topics[self.topic_name].get_simple_consumer()
        try:
            messages = [consumer.consume() for _ in xrange(1000)]
            self.assertEquals(len(messages), 1000)
        finally:
            consumer.stop()

    def test_offset_commit(self):
        consumer = self.client.topics[self.topic_name].get_simple_consumer('test_offset_commit')
        try:
            [consumer.consume() for _ in xrange(100)]
            consumer.commit_offsets()
            res = consumer.fetch_offsets()
            offset_sum = sum(r[1].offset for r in res if r[1].offset >= 0)
            self.assertEquals(offset_sum, 99)  # 0-indexed
        finally:
            consumer.stop()


    def test_offset_resume(self):
        consumer = self.client.topics[self.topic_name].get_simple_consumer('test_offset_resume')
        try:
            [consumer.consume() for _ in xrange(100)]
            consumer.commit_offsets()
        finally:
            consumer.stop()

        consumer = self.client.topics[self.topic_name].get_simple_consumer('test_offset_resume')
        try:
            res = consumer.fetch_offsets()
            offset_sum = sum(p.last_offset_consumed
                             for p in consumer.partitions
                             if p.last_offset_consumed >= 0)
            self.assertEquals(offset_sum, 99)  # 0-indexed
        finally:
            consumer.stop()


class TestOwnedPartition(unittest2.TestCase):
    def test_partition_saves_offset(self):
        msgval = "test"
        op = OwnedPartition(None)

        message = mock.Mock()
        message.value = msgval
        message.offset = 20

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 1)
        ret_message = op.consume()
        self.assertEqual(op.last_offset_consumed, message.offset)
        self.assertEqual(op.next_offset, message.offset + 1)
        self.assertNotEqual(ret_message, None)
        self.assertEqual(ret_message.value, msgval)

    def test_partition_rejects_old_message(self):
        last_offset = 400
        op = OwnedPartition(None)
        op.last_offset_consumed = last_offset

        message = mock.Mock()
        message.value = "test"
        message.offset = 20

        op.enqueue_messages([message])
        self.assertEqual(op.message_count, 0)
        op.consume()
        self.assertEqual(op.last_offset_consumed, last_offset)

    def test_partition_consume_empty_queue(self):
        op = OwnedPartition(None)

        message = op.consume()
        self.assertEqual(message, None)

    def test_partition_offset_commit_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)
        op.last_offset_consumed = 200

        rqtime = int(time.time())
        request = op.build_offset_commit_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)
        self.assertEqual(request.offset, op.last_offset_consumed)
        self.assertEqual(request.metadata, 'pykafka')

    def test_partition_offset_fetch_request(self):
        topic = mock.Mock()
        topic.name = "test_topic"
        partition = mock.Mock()
        partition.topic = topic
        partition.id = 12345

        op = OwnedPartition(partition)

        request = op.build_offset_fetch_request()

        self.assertEqual(request.topic_name, topic.name)
        self.assertEqual(request.partition_id, partition.id)

    def test_partition_offset_counters(self):
        res = mock.Mock()
        res.offset = 400

        op = OwnedPartition(None)
        op.set_offset(res.offset)

        self.assertEqual(op.last_offset_consumed, res.offset)
        self.assertEqual(op.next_offset, res.offset + 1)


if __name__ == "__main__":
    unittest2.main()



import unittest2

from pykafka import KafkaClient
from pykafka.test.utils import get_cluster, stop_cluster


class ProducerIntegrationTests(unittest2.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.topic_name = 'test-data'
        cls.kafka.create_topic(cls.topic_name, 3, 2)
        cls.client = KafkaClient(cls.kafka.brokers)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_produce(self):
        try:
            prod = self.client.topics[self.topic_name].get_producer(batch_size=5)
            prod.produce(['msg test'])

            consumer = self.client.topics[self.topic_name].get_balanced_consumer('test_consume', zookeeper_connect=self.kafka.zookeeper)
            messages = [consumer.consume() for i in xrange(1)]
            self.assertTrue(len(messages) == 1)
        finally:
            consumer.stop()

if __name__ == "__main__":
    unittest2.main()



from __future__ import division
"""
Author: Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["SimpleConsumer"]
import itertools
import logging
import time
import threading
from collections import defaultdict
from Queue import Queue, Empty

import base
from .common import OffsetType
from .utils.compat import Semaphore
from .exceptions import (OffsetOutOfRangeError, UnknownTopicOrPartition,
                         OffsetMetadataTooLarge, OffsetsLoadInProgress,
                         NotCoordinatorForConsumer, SocketDisconnectedError,
                         ConsumerStoppedException, ERROR_CODES)
from .protocol import (PartitionFetchRequest, PartitionOffsetCommitRequest,
                       PartitionOffsetFetchRequest, PartitionOffsetRequest)
from .utils.error_handlers import handle_partition_responses, raise_error


log = logging.getLogger(__name__)


class SimpleConsumer(base.BaseSimpleConsumer):
    """
    A non-balancing consumer for Kafka
    """
    def __init__(self,
                 topic,
                 cluster,
                 consumer_group=None,
                 partitions=None,
                 fetch_message_max_bytes=1024 * 1024,
                 num_consumer_fetchers=1,
                 auto_commit_enable=False,
                 auto_commit_interval_ms=60 * 1000,
                 queued_max_messages=2000,
                 fetch_min_bytes=1,
                 fetch_wait_max_ms=100,
                 refresh_leader_backoff_ms=200,
                 offsets_channel_backoff_ms=1000,
                 offsets_commit_max_retries=5,
                 auto_offset_reset=OffsetType.LATEST,
                 consumer_timeout_ms=-1,
                 auto_start=True,
                 reset_offset_on_start=False):
        """Create a SimpleConsumer.

        Settings and default values are taken from the Scala
        consumer implementation.  Consumer group is included
        because it's necessary for offset management, but doesn't imply
        that this is a balancing consumer. Use a BalancedConsumer for
        that.

        :param topic: The topic this consumer should consume
        :type topic: :class:`pykafka.topic.Topic`
        :param cluster: The cluster to which this consumer should connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param consumer_group: The name of the consumer group to join
        :type consumer_group: str
        :param partitions: Existing partitions to which to connect
        :type partitions: Iterable of :class:`pykafka.partition.Partition`
        :param fetch_message_max_bytes: The number of bytes of messages to
            attempt to fetch
        :type fetch_message_max_bytes: int
        :param num_consumer_fetchers: The number of workers used to fetch data
        :type num_consumer_fetchers: int
        :param auto_commit_enable: If true, periodically commit to kafka the
            offset of messages already fetched by this consumer.
        :type auto_commit_enable: bool
        :param auto_commit_interval_ms: The frequency (in milliseconds) at which the
            consumer offsets are committed to kafka
        :type auto_commit_interval_ms: int
        :param queued_max_messages: Maximum number of messages buffered for
            consumption
        :type queued_max_messages: int
        :param fetch_min_bytes: The minimum amount of data (in bytes) the server
            should return for a fetch request. If insufficient data is available
            the request will block.
        :type fetch_min_bytes: int
        :param fetch_wait_max_ms: The maximum amount of time (in milliseconds)
            the server will block before answering the fetch request if there
            isn't sufficient data to immediately satisfy fetch_min_bytes.
        :type fetch_wait_max_ms: int
        :param refresh_leader_backoff_ms: Backoff time (in milliseconds) to
            refresh the leader of a partition after it loses the current leader.
        :type refresh_leader_backoff_ms: int
        :param offsets_channel_backoff_ms: Backoff time (in milliseconds) to
            retry offset commits/fetches
        :type offsets_channel_backoff_ms: int
        :param offsets_commit_max_retries: Retry the offset commit up to this
            many times on failure.
        :type offsets_commit_max_retries: int
        :param auto_offset_reset: What to do if an offset is out of range. This
            setting indicates how to reset the consumer's internal offset
            counter when an OffsetOutOfRangeError is encountered.
        :type auto_offset_reset: :class:`pykafka.common.OffsetType`
        :param consumer_timeout_ms: Amount of time (in milliseconds) the
            consumer may spend without messages available for consumption
            before raising an error.
        :type consumer_timeout_ms: int
        :param auto_start: Whether the consumer should begin communicating
            with kafka after __init__ is complete. If false, communication
            can be started with `start()`.
        :type auto_start: bool
        :param reset_offset_on_start: Whether the consumer should reset its
            internal offset counter to `self._auto_offset_reset` immediately
            upon starting up
        :type reset_offset_on_start: bool
        """
        self._cluster = cluster
        self._consumer_group = consumer_group
        self._topic = topic
        self._fetch_message_max_bytes = fetch_message_max_bytes
        self._fetch_min_bytes = fetch_min_bytes
        self._queued_max_messages = queued_max_messages
        self._num_consumer_fetchers = num_consumer_fetchers
        self._fetch_wait_max_ms = fetch_wait_max_ms
        self._consumer_timeout_ms = consumer_timeout_ms
        self._offsets_channel_backoff_ms = offsets_channel_backoff_ms
        self._auto_offset_reset = auto_offset_reset
        self._offsets_commit_max_retries = offsets_commit_max_retries
        # not directly configurable
        self._offsets_fetch_max_retries = offsets_commit_max_retries
        self._offsets_reset_max_retries = offsets_commit_max_retries
        self._auto_start = auto_start
        self._reset_offset_on_start = reset_offset_on_start

        # incremented for any message arrival from any partition
        # the initial value is 0 (no messages waiting)
        self._messages_arrived = Semaphore(value=0)

        self._auto_commit_enable = auto_commit_enable
        self._auto_commit_interval_ms = auto_commit_interval_ms
        self._last_auto_commit = time.time()

        self._discover_offset_manager()

        if partitions:
            self._partitions = {OwnedPartition(p, self._messages_arrived): p
                                for p in partitions}
        else:
            self._partitions = {OwnedPartition(p, self._messages_arrived):
                                topic.partitions[k]
                                for k, p in topic.partitions.iteritems()}
        self._partitions_by_id = {p.partition.id: p
                                  for p in self._partitions.iterkeys()}
        # Organize partitions by leader for efficient queries
        self._partitions_by_leader = defaultdict(list)
        for p in self._partitions.iterkeys():
            self._partitions_by_leader[p.partition.leader].append(p)
        self.partition_cycle = itertools.cycle(self._partitions.keys())

        self._default_error_handlers = self._build_default_error_handlers()

        self._running = False
        if self._auto_start:
            self.start()

    def __repr__(self):
        return "<{}.{} at {} (consumer_group={})>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self._consumer_group
        )

    def start(self):
        """Begin communicating with Kafka, including setting up worker threads

        Fetches offsets, starts an offset autocommitter worker pool, and
        starts a message fetcher worker pool.
        """
        self._running = True

        # Figure out which offset wer're starting on
        if self._reset_offset_on_start:
            self._reset_offsets()
            # make sure the reset is saved in kafka before it can rebalance
            self.commit_offsets()
        elif self._consumer_group is not None:
            self.fetch_offsets()

        self._fetch_workers = self._setup_fetch_workers()

        if self._auto_commit_enable:
            self._autocommit_worker_thread = self._setup_autocommit_worker()

    def _build_default_error_handlers(self):
        """Set up the error handlers to use for partition errors."""
        def _handle_OffsetOutOfRangeError(parts):
            self._reset_offsets([owned_partition
                                 for owned_partition, pres in parts])

        def _handle_NotCoordinatorForConsumer(parts):
            self._discover_offset_manager()

        return {
            UnknownTopicOrPartition.ERROR_CODE: lambda p: raise_error(UnknownTopicOrPartition),
            OffsetOutOfRangeError.ERROR_CODE: _handle_OffsetOutOfRangeError,
            OffsetMetadataTooLarge.ERROR_CODE: lambda p: raise_error(OffsetMetadataTooLarge),
            NotCoordinatorForConsumer.ERROR_CODE: _handle_NotCoordinatorForConsumer
        }

    def _discover_offset_manager(self):
        """Set the offset manager for this consumer.

        If a consumer group is not supplied to __init__, this method does nothing
        """
        if self._consumer_group is not None:
            self._offset_manager = self._cluster.get_offset_manager(self._consumer_group)

    @property
    def topic(self):
        """The topic this consumer consumes"""
        return self._topic

    @property
    def partitions(self):
        """A list of the partitions that this consumer consumes"""
        return self._partitions

    def __del__(self):
        """Stop consumption and workers when object is deleted"""
        self.stop()

    def stop(self):
        """Flag all running workers for deletion."""
        self._running = False

    def _setup_autocommit_worker(self):
        """Start the autocommitter thread"""
        def autocommitter():
            while True:
                if not self._running:
                    break
                if self._auto_commit_enable:
                    self._auto_commit()
                time.sleep(self._auto_commit_interval_ms / 1000)
            log.debug("Autocommitter thread exiting")
        log.debug("Starting autocommitter thread")
        return self._cluster.handler.spawn(autocommitter)

    def _setup_fetch_workers(self):
        """Start the fetcher threads"""
        def fetcher():
            while True:
                if not self._running:
                    break
                self.fetch()
                time.sleep(.0001)
            log.debug("Fetcher thread exiting")
        log.info("Starting %s fetcher threads", self._num_consumer_fetchers)
        return [self._cluster.handler.spawn(fetcher)
                for i in xrange(self._num_consumer_fetchers)]

    def __iter__(self):
        """Yield an infinite stream of messages until the consumer times out"""
        while True:
            message = self.consume(block=True)
            if not message:
                raise StopIteration
            yield message

    def consume(self, block=True):
        """Get one message from the consumer.

        :param block: Whether to block while waiting for a message
        :type block: bool
        """
        timeout = None
        if block:
            if self._consumer_timeout_ms > 0:
                timeout = float(self._consumer_timeout_ms) / 1000
            else:
                timeout = 1.0

        while True:
            if self._messages_arrived.acquire(blocking=block, timeout=timeout):
                # by passing through this semaphore, we know that at
                # least one message is waiting in some queue.
                message = None
                while not message:
                    owned_partition = self.partition_cycle.next()
                    message = owned_partition.consume()
                return message
            else:
                if not self._running:
                    raise ConsumerStoppedException()
                elif not block or self._consumer_timeout_ms > 0:
                    return None

    def _auto_commit(self):
        """Commit offsets only if it's time to do so"""
        if not self._auto_commit_enable or self._auto_commit_interval_ms == 0:
            return

        if (time.time() - self._last_auto_commit) * 1000.0 >= self._auto_commit_interval_ms:
            log.info("Autocommitting consumer offset for consumer group %s and topic %s",
                     self._consumer_group, self._topic.name)
            self.commit_offsets()
            self._last_auto_commit = time.time()

    def commit_offsets(self):
        """Commit offsets for this consumer's partitions

        Uses the offset commit/fetch API
        """
        if not self._consumer_group:
            raise Exception("consumer group must be specified to commit offsets")

        reqs = [p.build_offset_commit_request() for p in self._partitions.keys()]
        log.debug("Committing offsets for %d partitions to broker id %s", len(reqs),
                  self._offset_manager.id)
        for i in xrange(self._offsets_commit_max_retries):
            if i > 0:
                log.debug("Retrying")
            time.sleep(i * (self._offsets_channel_backoff_ms / 1000))

            response = self._offset_manager.commit_consumer_group_offsets(
                self._consumer_group, 1, 'pykafka', reqs)
            parts_by_error = handle_partition_responses(
                response,
                self._default_error_handlers,
                partitions_by_id=self._partitions_by_id)
            if len(parts_by_error) == 1 and 0 in parts_by_error:
                break
            log.error("Error committing offsets for topic %s (errors: %s)",
                      self._topic.name,
                      {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                       for err, parts in parts_by_error.iteritems()})

            # retry only the partitions that errored
            if 0 in parts_by_error:
                parts_by_error.pop(0)
            errored_partitions = [op for code, err_group in parts_by_error.iteritems()
                                  for op, res in err_group]
            reqs = [p.build_offset_commit_request() for p in errored_partitions]

    def fetch_offsets(self):
        """Fetch offsets for this consumer's topic

        Uses the offset commit/fetch API

        :return: List of (id, :class:`pykafka.protocol.OffsetFetchPartitionResponse`)
            tuples
        """
        if not self._consumer_group:
            raise Exception("consumer group must be specified to fetch offsets")

        def _handle_success(parts):
            for owned_partition, pres in parts:
                log.debug("Set offset for partition %s to %s",
                          owned_partition.partition.id,
                          pres.offset)
                owned_partition.set_offset(pres.offset)

        reqs = [p.build_offset_fetch_request() for p in self._partitions.keys()]
        success_responses = []

        log.debug("Fetching offsets for %d partitions from broker id %s", len(reqs),
                  self._offset_manager.id)

        for i in xrange(self._offsets_fetch_max_retries):
            if i > 0:
                log.debug("Retrying offset fetch")

            res = self._offset_manager.fetch_consumer_group_offsets(self._consumer_group, reqs)
            parts_by_error = handle_partition_responses(
                res,
                self._default_error_handlers,
                success_handler=_handle_success,
                partitions_by_id=self._partitions_by_id)

            success_responses.extend([(op.partition.id, r)
                                      for op, r in parts_by_error.get(0, [])])
            if len(parts_by_error) == 1 and 0 in parts_by_error:
                return success_responses
            log.error("Error fetching offsets for topic %s (errors: %s)",
                      self._topic.name,
                      {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                       for err, parts in parts_by_error.iteritems()})

            time.sleep(i * (self._offsets_channel_backoff_ms / 1000))

            # retry only specific error responses
            to_retry = []
            to_retry.extend(parts_by_error.get(OffsetsLoadInProgress.ERROR_CODE, []))
            to_retry.extend(parts_by_error.get(NotCoordinatorForConsumer.ERROR_CODE, []))
            reqs = [p.build_offset_fetch_request() for p, _ in to_retry]

    def _reset_offsets(self, partitions=None):
        """Reset offsets after an error

        Issue an OffsetRequest for each partition and set the appropriate
        returned offset in the OwnedPartition per self._auto_offset_reset

        :param partitions: the partitions for which to reset offsets
        :type partitions: Iterable of
            :class:`pykafka.simpleconsumer.OwnedPartition`
        """
        def _handle_success(parts):
            for owned_partition, pres in parts:
                # offset_latest requests return the next offset to consume,
                # so account for this here by passing offset - 1
                owned_partition.set_offset(pres.offset[0] - 1)

        if partitions is None:
            partitions = self._partitions.keys()

        log.info("Resetting offsets for %s partitions", len(list(partitions)))

        for i in xrange(self._offsets_reset_max_retries):
            # group partitions by leader
            by_leader = defaultdict(list)
            for p in partitions:
                by_leader[p.partition.leader].append(p)

            # get valid offset ranges for each partition
            for broker, owned_partitions in by_leader.iteritems():
                reqs = [owned_partition.build_offset_request(self._auto_offset_reset)
                        for owned_partition in owned_partitions]
                response = broker.request_offset_limits(reqs)
                parts_by_error = handle_partition_responses(
                    response,
                    self._default_error_handlers,
                    success_handler=_handle_success,
                    partitions_by_id=self._partitions_by_id)

                if len(parts_by_error) == 1 and 0 in parts_by_error:
                    break
                log.error("Error resetting offsets for topic %s (errors: %s)",
                          self._topic.name,
                          {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                           for err, parts in parts_by_error.iteritems()})

                time.sleep(i * (self._offsets_channel_backoff_ms / 1000))

                if 0 in parts_by_error:
                    parts_by_error.pop(0)
                partitions = []
                partitions.extend(
                    [part for errcode, parts in parts_by_error.iteritems()
                     for part in parts])

    def fetch(self):
        """Fetch new messages for all partitions

        Create a FetchRequest for each broker and send it. Enqueue each of the
        returned messages in the approprate OwnedPartition.
        """
        def _handle_success(parts):
            for owned_partition, pres in parts:
                if len(pres.messages) > 0:
                    log.debug("Fetched %s messages for partition %s",
                              len(pres.messages), owned_partition.partition.id)
                    owned_partition.enqueue_messages(pres.messages)
                    log.debug("Partition %s queue holds %s messages",
                              owned_partition.partition.id,
                              owned_partition.message_count)

        for broker, owned_partitions in self._partitions_by_leader.iteritems():
            partition_reqs = {}
            for owned_partition in owned_partitions:
                # attempt to acquire lock, just pass if we can't
                if owned_partition.lock.acquire(False):
                    partition_reqs[owned_partition] = None
                    if owned_partition.message_count < self._queued_max_messages:
                        fetch_req = owned_partition.build_fetch_request(
                            self._fetch_message_max_bytes)
                        partition_reqs[owned_partition] = fetch_req
                    else:
                        log.debug("Partition %s above max queued count (queue has %d)",
                                  owned_partition.partition.id,
                                  owned_partition.message_count)
            if partition_reqs:
                try:
                    response = broker.fetch_messages(
                        [a for a in partition_reqs.itervalues() if a],
                        timeout=self._fetch_wait_max_ms,
                        min_bytes=self._fetch_min_bytes
                    )
                except SocketDisconnectedError:
                    # If the broker dies while we're supposed to stop,
                    # it's fine, and probably an integration test.
                    if not self._running:
                        return
                    else:
                        raise
                handle_partition_responses(
                    response,
                    self._default_error_handlers,
                    success_handler=_handle_success,
                    partitions_by_id=self._partitions_by_id)
            for owned_partition in partition_reqs.iterkeys():
                owned_partition.lock.release()


class OwnedPartition(object):
    """A partition that is owned by a SimpleConsumer.

    Used to keep track of offsets and the internal message queue.
    """

    def __init__(self,
                 partition,
                 semaphore=None):
        """
        :param partition: The partition to hold
        :type partition: :class:`pykafka.partition.Partition`
        :param semaphore: A Semaphore that counts available messages and
            facilitates non-busy blocking
        :type semaphore: :class:`pykafka.utils.compat.Semaphore`
        """
        self.partition = partition
        self._messages = Queue()
        self._messages_arrived = semaphore
        self.last_offset_consumed = 0
        self.next_offset = 0
        self.lock = threading.Lock()

    @property
    def message_count(self):
        """Count of messages currently in this partition's internal queue"""
        return self._messages.qsize()

    def set_offset(self, last_offset_consumed):
        """Set the internal offset counters

        :param last_offset_consumed: The last committed offset for this
            partition
        :type last_offset_consumed: int
        """
        self.last_offset_consumed = last_offset_consumed
        self.next_offset = last_offset_consumed + 1

    def build_offset_request(self, auto_offset_reset):
        """Create a :class:`pykafka.protocol.PartitionOffsetRequest` for this
            partition

        :param auto_offset_reset: What to do if an offset is out of range. This
            setting indicates how to reset the consumer's internal offset
            counter when an OffsetOutOfRangeError is encountered.
        :type auto_offset_reset: :class:`pykafka.common.OffsetType`
        """
        return PartitionOffsetRequest(
            self.partition.topic.name, self.partition.id,
            auto_offset_reset, 1)

    def build_fetch_request(self, max_bytes):
        """Create a :class:`pykafka.protocol.FetchPartitionRequest` for this
            partition.

        :param max_bytes: The number of bytes of messages to
            attempt to fetch
        :type max_bytes: int
        """
        return PartitionFetchRequest(
            self.partition.topic.name, self.partition.id,
            self.next_offset, max_bytes)

    def build_offset_commit_request(self):
        """Create a :class:`pykafka.protocol.PartitionOffsetCommitRequest`
            for this partition
        """
        return PartitionOffsetCommitRequest(
            self.partition.topic.name,
            self.partition.id,
            self.last_offset_consumed,
            int(time.time() * 1000),
            'pykafka'
        )

    def build_offset_fetch_request(self):
        """Create a PartitionOffsetFetchRequest for this partition
        """
        return PartitionOffsetFetchRequest(
            self.partition.topic.name,
            self.partition.id
        )

    def consume(self):
        """Get a single message from this partition"""
        try:
            message = self._messages.get_nowait()
            self.last_offset_consumed = message.offset
            return message
        except Empty:
            return None

    def enqueue_messages(self, messages):
        """Put a set of messages into the internal message queue

        :param messages: The messages to enqueue
        :type messages: Iterable of :class:`pykafka.common.Message`
        """
        for message in messages:
            if message.offset < self.last_offset_consumed:
                log.debug("Skipping enqueue for offset (%s) "
                          "less than last_offset_consumed (%s)",
                          message.offset, self.last_offset_consumed)
                continue
            self._messages.put(message)
            self.next_offset = message.offset + 1

            if self._messages_arrived is not None:
                self._messages_arrived.release()
